"""
run this before using this library:
ipython -i cpdctrl_interactive.py

always records iv-t curves
    i-data -> smua.nvbuffer1
    v-data -> smua.nvbuffer2
"""
version = "0.1"

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from datetime import datetime as dtime
from sys import exit
from time import sleep
from os import path, makedirs
import pickle as pkl
import json
import atexit

import argparse

if __name__ == "__main__":
    import sys
    if __package__ is None:
        # make relative imports work as described here: https://peps.python.org/pep-0366/#proposed-change
        __package__ = "cpdctrl"
        from os import path
        filepath = path.realpath(path.abspath(__file__))
        sys.path.insert(0, path.dirname(path.dirname(filepath)))
    parser = argparse.ArgumentParser(
        prog="cpdctrl",
        description="measure voltage using a Keithley SMU",
    )
    backend_group = parser.add_mutually_exclusive_group(required=False)
    backend_group.add_argument("-k", "--keithley", action="store_true")
    # backend_group.add_argument("-t", "--testing", action='store_true')
    parser.add_argument("-c", "--config", action="store", help="alternate path to config file")
    args = vars(parser.parse_args())


from .voltage_measurement_device.base import VoltageMeasurementDevice
from .voltage_measurement_device.impl import keithley2700 as _volt
from .led_control_device.base import LedControlDevice
from .led_control_device.impl import thorlabs_ledd1b as _led
from .led_script import LedScript

from .measurement import Measurement
from .utility.data import DataCollector



from .utility import data as _data
from .utility.data import load_dataframe, plot_cpd_data
from .utility import file_io
from .update_funcs import _Monitor, _update_print

config_path = path.expanduser("~/.config/cpdctrl.json")

_runtime_vars = {
    "last-measurement": ""
}

settings = {
    "datadir":      path.expanduser("~/data"),
    "name":         "measurement",
    "interval":     0.02,
    "beep":         True,
}

test = False

# global variable for the instrument/client returned by pyvisa/bleak
dev: VoltageMeasurementDevice|None = None
led: LedControlDevice|None = None
data = DataCollector(data_path=settings["datadir"], data_name="interactive", dirname="interactive_test", dir_exists_is_ok=True)
t0 = 0

def monitor(interval=None, max_measurements=None, max_points_shown=160):
    """
    Monitor the voltage with matplotlib.

    @details:
        - Resets the buffers
        - Opens a matplotlib window and takes measurements depending on settings["interval"]
        - Waits for the user to press a key
        Uses python's time.sleep() for waiting the interval, which is not very precise. Use measure_count for better precision.
        You can take the data from the buffer afterwards, using save_csv.
    @param max_points_shown : how many points should be shown at once. None means infinite
    @param max_measurements : maximum number of measurements. None means infinite
    """
    global _runtime_vars
    _runtime_vars["last_measurement"] = dtime.now().isoformat()
    if not interval: interval = settings["interval"]
    print(f"Starting measurement with:\n\tinterval = {interval}s\nUse <C-c> to stop. Save the data using 'save_csv()' afterwards.")
    plt_monitor = _Monitor(use_print=True, max_points_shown=max_points_shown)
    data.clear()
    def update_func(i, t, v):
        global t0
        if i == 0:
            t0 = t
        t -= t0
        data.add_data(i, t, v, 0)
        plt_monitor.update(i, t, v)
        # update_led()
    dev.measure(interval=interval, max_measurements=max_measurements, update_func=update_func)

def measure(interval=None, max_measurements=None):
    """
    Measure voltages

    @details:
        - Resets the buffers
        - Measure voltages
        - Waits for the user to press a key
        Uses python's time.sleep() for waiting the interval, which is not very precise. Use measure_count for better precision.
        You can take the data from the buffer afterwards, using save_csv.
    @param max_measurements : maximum number of measurements. None means infinite
    """
    global _runtime_vars
    if not interval: interval = settings["interval"]
    _runtime_vars["last_measurement"] = dtime.now().isoformat()
    print(f"Starting measurement with:\n\tinterval = {interval}s\nUse <C-c> to stop. Save the data using 'save_csv()' afterwards.")
    update_func = _update_print
    dev.measure(interval=interval, max_measurements=max_measurements, update_func=update_func)
    

def save_csv():
    """
    Saves the contents of nvbuffer1 as .csv
    The settings 'datadir' and 'name' are used for determining the filepath:
    'datadir/nameXXX.csv', where XXX is the number of files that exist in datadir with the same name.
    """
    df = get_dataframe()
    filename = settings["datadir"] + "/" + df.basename + ".csv"
    df.to_csv(filename, index=False, header=True)
    print(f"Saved as '{filename}'")


def save_pickle():
    """
    Saves the contents of nvbuffer1 as .pkl
    The settings 'datadir' and 'name' are used for determining the filepath:
    'datadir/nameXXX.pkl', where XXX is the number of files that exist in datadir with the same name.
    """
    df = get_dataframe()
    filename = settings["datadir"] + "/" + df.basename + ".pkl"
    df.to_pickle(filename)
    print(f"Saved as '{filename}'")


def run_script(script_path):
    """
    Run a lua script on the Keithley device
    @param script_path : relative or absolute path to the .lua script
    """
    global k, settings
    if test:
        print("run_script: Test mode enabled, ignoring call to run_script")
    else:
        _backend.run_lua(dev, script_path=script_path)


def set(setting, value):
    global settings, config_path
    if setting in settings:
        if type(value) != type(settings[setting]):
            print(f"set: setting '{setting}' currently holds a value of type '{type(settings[setting])}'")
            return
    settings[setting] = value

def name(s:str):
    global settings
    settings["name"] = s

def save_settings():
    with open(config_path, "w") as file:
        json.dump(settings, file, indent=4)

def load_settings():
    global settings, config_path
    with open(config_path, "r") as file:
        settings = json.load(file)
    settings["datadir"] = path.expanduser(settings["datadir"])  # replace ~

def help(topic=None):
    if topic == None:
        print("""
Functions:
    measure         [kat] - take measurements
    monitor         [kat] - take measurements with live monitoring in a matplotlib window
    measure_count   [kat] - take a fixed number of measurements
    monitor_count   [kat] - take a fixed number of measurements with live monitoring in a matplotlib window
    repeat          [kat] - measure and save to csv multiple times
    get_dataframe   [kat] - return device internal buffer as pandas dataframe
    save_csv        [kat] - save the last measurement as csv file
    save_pickle     [kat] - save the last measurement as pickled pandas dataframe
    load_dataframe  [kat] - load a pandas dataframe from csv or pickle
    run_script      [k  ] - run a lua script on the Keithely device
Run 'help(function)' to see more information on a function

Available topics:
    imports
    device
    settings
Run 'help("topic")' to see more information on a topic""")

    elif topic in [settings, "settings"]:
        print("""Settings:
    name: str       - name of the measurement, determines filename of 'save_csv'
    datadir: str    - output directory for the csv files
    interval: int   - interval (inverse frequency) of the measurements, in seconds
    beep: bool      - wether the device should beep or not

Functions:
    name("<name>")         - short for set("name", "<name>")
    set("setting", value)  - set a setting to a value
    save_settings()        - store the settings as "cpdctrl.json" in the working directory
    load_settings()        - load settings from a file
    The global variable 'config_path' determines the path used by save/load_settings. Use -c '<path>' to set another path.
    The serach path is:
        <working-dir>/cpdctrl.json
        $XDG_CONFIG_HOME/cpdctrl.json
        ~/.config/cpdctrl.json
              """)
    elif topic == "imports":
        print("""Imports:
    numpy as np
    pandas as pd
    matplotlib.pyplot as plt
    os.path """)
    elif topic == "device":
        print("""Device:
    keithley backend:
        The opened pyvisa resource (deveithley device) is the global variable 'dev'.
        You can interact using pyvisa functions, such as
        k.write("command"), k.query("command") etc. to interact with the device.
    arduino backend:
        The Arduino will be avaiable as BleakClient using the global variable 'dev'.  """)
    else:
        print(topic.__doc__)


def init():
    global dev, led, settings, config_path
    print(r"""                  .___      __         .__   
  ____ ______   __| _/_____/  |________|  |  
_/ ___\\____ \ / __ |/ ___\   __\_  __ \  |  
\  \___|  |_> > /_/ \  \___|  |  |  | \/  |__
 \___  >   __/\____ |\___  >__|  |__|  |____/
     \/|__|        \/    \/                  """ + f"""{version}
Interactive Shell for CPD measurements with Keithley 2700B
---
Enter 'help()' for a list of commands""")
    from os import environ
    if path.isfile("cpdctrl.json"):
        config_path = "cpdctrl.json"
    elif 'XDG_CONFIG_HOME' in environ.keys():
        # and path.isfile(environ["XDG_CONFIG_HOME"] + "/cpdctrl.json"):
        config_path = environ["XDG_CONFIG_HOME"] + "/cpdctrl.json"
    else:
        config_path = path.expanduser("~/.config/cpdctrl.json")
    if args["config"]:
        config_path = args["config"]

 
    if not path.isdir(path.dirname(config_path)):
        makedirs(path.dirname(config_path))

    if path.isfile(config_path):
        load_settings()

    if not path.isdir(settings["datadir"]):
        makedirs(settings["datadir"])

    try:
        dev = _volt.init("GPIB0::22::INSTR")
        led = _led.LEDD1B()
    except Exception as e:
        print(e)
        exit(1)
    # atexit.register(_backend.exit, dev)


if __name__ == "__main__":
    init()
